from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Callable, Tuple, Any, Optional
from enum import Enum


class AstNode(ABC):
    @property
    def childs(self) -> Tuple['AstNode', ...]:
        return ()

    @abstractmethod
    def __str__(self) -> str:
        pass

    @property
    def tree(self) -> [str, ...]:
        res = [str(self)]
        childs = self.childs
        for i, child in enumerate(childs):
            ch0, ch = '├', '│'
            if i == len(childs) - 1:
                ch0, ch = '└', ' '
            res.extend(((ch0 if j == 0 else ch) + ' ' + s for j, s in enumerate(child.tree)))
        return res

    def visit(self, func: Callable[['AstNode'], None]) -> None:
        func(self)
        map(func, self.childs)

    def __getitem__(self, index):
        return self.childs[index] if index < len(self.childs) else None

    def execute(self, intepreter, context):
        return intepreter.execute(self, context)


class ExprNode(AstNode):
    pass


class StmtNode(AstNode):
    pass


class LiteralNode(ExprNode):
    def __init__(self, value: Any):
        super().__init__()
        self.value = value
        if value is None:
            self.value = True
            return
        if isinstance(value, str) and (value.startswith('"') or value.startswith('\'')):
            self.value = value[1:-1]
        with suppress(Exception):
            self.value = int(value)
            return
        with suppress(Exception):
            self.value = float(value)
            return
        if value in ('true', 'false'):
            self.value = value == 'true'

    def __str__(self) -> str:
        if isinstance(self.value, bool):
            return str(self.value).lower()
        if isinstance(self.value, str):
            return f'"{self.value}"'
        return str(self.value)


class IdentNode(ExprNode):
    def __init__(self, name: str):
        super().__init__()
        self.name = str(name)

    def __str__(self) -> str:
        return str(self.name)


class UnOp(Enum):
    NOT = 'not'
    MINUS = '-'


class BinOp(Enum):
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    GT = '>'
    GE = '>='
    LT = '<'
    LE = '<='
    EQ = '='
    NE = '<>'
    AND = 'and'
    OR = 'or'
    CONC = '||'


class UnOpNode(ExprNode):
    def __init__(self, op: UnOp, arg: ExprNode):
        super().__init__()
        self.op = op
        self.arg = arg

    @property
    def childs(self) -> Tuple[ExprNode]:
        return self.arg,

    def __str__(self) -> str:
        return str(self.op.value)


class BinOpNode(ExprNode):
    def __init__(self, op: BinOp, arg1: ExprNode, arg2: ExprNode):
        super().__init__()
        self.op = op
        self.arg1 = arg1
        self.arg2 = arg2

    @property
    def childs(self) -> Tuple[ExprNode, ExprNode]:
        return self.arg1, self.arg2

    def __str__(self) -> str:
        return str(self.op.value)


class FuncCallNode(ExprNode):
    def __init__(self, func: IdentNode, *params: ExprNode):
        super().__init__()
        self.func = func
        self.params = params

    @property
    def childs(self) -> Tuple[ExprNode]:
        return self.params

    def __str__(self) -> str:
        return f'{self.func}()'


class ExprListNode(ExprNode):
    def __init__(self, *exprs: ExprNode):
        super().__init__()
        self.exprs = exprs

    @property
    def childs(self) -> Tuple[ExprNode]:
        return self.exprs

    def __str__(self) -> str:
        return '...'


class SelectNode(ExprNode):
    def __init__(self, select: ExprListNode, tables: IdentNode, where: ExprNode, order: ExprListNode):
        super().__init__()
        self.select = select
        self.tables = tables
        self.where = where
        self.order = order

    @property
    def childs(self) -> Tuple[AstNode]:
        return self.select, self.tables, self.where, self.order

    def __str__(self) -> str:
        return 'select'
