import re
from contextlib import suppress
from numbers import Number
from typing import List, Any, Dict

import visitor
from sql_ast import SelectNode, LiteralNode, ExprListNode, IdentNode, BinOpNode, BinOp, UnOpNode, UnOp


class Table:
    def __init__(self, col_names: List[str], rows: List[List[Any]]):
        self.col_names = col_names
        self.rows = rows

    def get(self, name: str, row_index) -> Any:
        return self.rows[row_index][self.col_names.index(name)]

    @property
    def is_scalar(self) -> Any:
        return len(rows) == 1 and len(rows[0]) == 1

    @property
    def scalar(self) -> Any:
        return self.rows[0][0]


class Db:
    def __init__(self, tables: Dict[str, Table]):
        self.tables = {}
        for name, table in tables.items():
            col_names = [n.upper() for n in table.col_names]
            table = Table(col_names, table.rows)
            self.tables[name.upper()] = table


class Context(Db):
    def __init__(self, db: Db):
        self.db = db
        self.query_tables = {}
        self.query_table_indexes = {}


class Intepreter:

    @visitor.on('AstNode')
    def execute(self, AstNode):
        """
        Нужен для работы модуля visitor (инициализации диспетчера)
        """
        pass

    @visitor.when(LiteralNode)
    def execute(self, node: LiteralNode, context: Context):
        return node.value

    @visitor.when(IdentNode)
    def execute(self, node: IdentNode, context: Context):
        alias, *name = re.split('\.', node.name.upper())
        if len(name) == 0:
            alias, name = None, alias
        else:
            name = name[0]
        if alias:
            if alias not in context.query_tables:
                raise Exception(f'Отсутствует таблица {alias}')
            table = context.query_tables[alias]
            row_index = context.query_table_indexes[alias]
        else:
            found_count = 0
            for table_name, t in context.query_tables.items():
                if name in t.col_names:
                    found_count += 1
                    table = t
                    row_index = context.query_table_indexes[table_name]
            if found_count == 0:
                raise Exception(f'Не найден столбец {name}')
            if found_count > 1:
                raise Exception(f'Cтолбец {name} определен неоднозначно')
        return table.get(name, row_index)

    @staticmethod
    def to_num(val: Any) -> Any:
        if isinstance(val, Number):
            return val
        with suppress(Exception):
            return int(val)
        with suppress(Exception):
            return float(val)
        raise Exception(f'{val} is not number')

    @visitor.when(UnOpNode)
    def execute(self, node: UnOpNode, context: Context):
        arg = node.arg.execute(self, context)
        if node.op == UnOp.MINUS:
            return - self.to_num(arg)
        if node.op == UnOp.NOT:
            return not arg

    @visitor.when(BinOpNode)
    def execute(self, node: BinOpNode, context: Context):
        arg1 = node.arg1.execute(self, context)
        arg2 = node.arg2.execute(self, context)
        if node.op == BinOp.ADD:
            return self.to_num(arg1) + self.to_num(arg2)
        if node.op == BinOp.SUB:
            return self.to_num(arg1) - self.to_num(arg2)
        if node.op == BinOp.MUL:
            return self.to_num(arg1) * self.to_num(arg2)
        if node.op == BinOp.DIV:
            return self.to_num(arg1) / self.to_num(arg2)
        if node.op == BinOp.CONC:
            return str(arg1) + str(arg2)
        if node.op == BinOp.GT:
            return arg1 > arg2
        if node.op == BinOp.GE:
            return arg1 >= arg2
        if node.op == BinOp.LT:
            return arg1 < arg2
        if node.op == BinOp.LE:
            return arg1 <= arg2
        if node.op == BinOp.LT:
            return arg1 < arg2
        if node.op == BinOp.EQ:
            return arg1 == arg2
        if node.op == BinOp.NE:
            return arg1 != arg2
        if node.op == BinOp.OR:
            return arg1 or arg2
        if node.op == BinOp.AND:
            return arg1 and arg2
        raise Exception(f'Operation {node.op} not implemented')

    @visitor.when(ExprListNode)
    def execute(self, node: ExprListNode, context: Context):
        return [ch.execute(self, context) for ch in node.exprs]

    @visitor.when(SelectNode)
    def execute(self, node: SelectNode, context: Context):
        name, alias = re.split('\s+', node.tables.name.upper())
        alias = alias[0] if len(alias) > 0 else None
        table = context.db.tables[name]
        context.query_tables[alias or name] = table
        rows = []
        orders = []
        j = 0
        for i, row in enumerate(table.rows):
            context.query_table_indexes[alias or table] = i
            where = node.where.execute(self, context)
            if where:
                rows.append(node.select.execute(self, context))
                order = node.order.execute(self, context)
                order.append(j)
                orders.append(order)
                j += 1
        orders.sort()
        result = []
        for o in orders:
            result.append(rows[o[-1]])
        return result
