from lark import Lark, Transformer
from sql_ast import *


parser = Lark('''
    %import common.NUMBER
    %import common.CNAME
    %import common.NEWLINE
    %import common.WS
    %import common.CPP_COMMENT
    %import common.C_COMMENT
    %import common._STRING_ESC_INNER
    
    ESCAPED_STRING : "'" _STRING_ESC_INNER "'"

    %ignore WS

    COMMENT: CPP_COMMENT | C_COMMENT
    %ignore COMMENT

    BOOL: "true" | "false"
    literal: NUMBER | ESCAPED_STRING | BOOL
    TABLE: CNAME (WS+ CNAME)?
    table_name: TABLE -> ident
    COLUMN: CNAME ("." CNAME)?
    ident: COLUMN

    ?group: literal
        | ident
        | func_call
        | "(" expr ")"
        
    ?un1: group
        | "-" group         -> minus

    ?mult: un1
        | mult "*" un1      -> mul
        | mult "/" un1      -> div

    ?add: mult
        | add "+" mult      -> add
        | add "-" mult      -> sub
        
    ?conc: add
        | add "||" add
        
    ?compare: conc
        | conc ">" conc       -> gt
        | conc ">=" conc      -> ge
        | conc "<" conc       -> lt
        | conc "<=" conc      -> le
        | conc "=" conc       -> eq
        | conc "<>" conc      -> ne
        
    ?and: compare
        | and "and" compare
        
    ?or: and
        | or "or" and
        
    func_call: ident "(" (expr ("," expr)*)? ")" 
        
    ?expr: or

    expr_list: expr ("," expr)*
    select: "select" expr_list "from" table_name ("where" expr)? ("order" "by" expr_list)?

    ?start: select
''', start="start")


class MelASTBuilder(Transformer):
    def _call_userfunc(self, tree, new_children=None):
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            return f(*children)

    def __getattr__(self, item):
        if isinstance(item, str) and item.upper() == item:
            return lambda x: x

        if item == 'true':
            return lambda *args: LiteralNode('true')

        if item in ('minus', 'not'):
            def get_un_op_node(*args):
                op = UnOp[item.upper()]
                return UnOpNode(op, *args)
            return get_un_op_node

        if item in ('mul', 'div', 'add', 'sub', 'conc',
                    'gt', 'ge', 'lt', 'le', 'eq', 'ne',
                    'and', 'or'):
            def get_bin_op_node(*args):
                op = BinOp[item.upper()]
                return BinOpNode(op, *args)
            return get_bin_op_node

        if item == 'select':
            def get_select_node(*args):
                where = LiteralNode('true')
                order = ExprListNode(LiteralNode('1'))
                if len(args) == 4:
                    where = args[2]
                    order = args[3]
                elif len(args) == 3:
                    if isinstance(args[2], ExprListNode):
                        order = args[2]
                    else:
                        where = args[2]
                return SelectNode(args[0], args[1], where, order)

            return get_select_node

        else:
            def get_node(*args):
                cls = eval(''.join(x.capitalize() or '_' for x in item.split('_')) + 'Node')
                return cls(*args)
            return get_node


def parse(prog: str) -> SelectNode:
    prog = parser.parse(str(prog))
    prog = MelASTBuilder().transform(prog)
    return prog
